/*
* Licensed to the Apache Software Foundation (ASF) under one
* or more contributor license agreements. See the NOTICE file
* distributed with this work for additional information
* regarding copyright ownership. The ASF licenses this file
* to you under the Apache License, Version 2.0 (the
* "License"); you may not use this file except in compliance
* with the License. You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing,
* software distributed under the License is distributed on an
* "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
* KIND, either express or implied. See the License for the
* specific language governing permissions and limitations
* under the License.
*/

package org.apache.taverna.workflowmodel;

import static org.apache.taverna.annotation.HierarchyRole.CHILD;

import java.util.List;

import org.apache.taverna.annotation.Annotated;
import org.apache.taverna.annotation.HierarchyTraversal;
import org.apache.taverna.invocation.InvocationContext;
import org.apache.taverna.lang.observer.Observable;
import org.apache.taverna.workflowmodel.processor.activity.Activity;
import org.apache.taverna.workflowmodel.processor.dispatch.DispatchStack;
import org.apache.taverna.workflowmodel.processor.iteration.IterationStrategyStack;

/**
 * A single node within the dataflow digraph, the Processor is the basic
 * functional unit within a Taverna workflow. It should also notify interested
 * observers when it has finished execution (including all iterations of the
 * processor).
 * 
 * @author Tom Oinn
 * @author Alex Nenadic
 */
@ControlBoundary
public interface Processor extends TokenProcessingEntity, Annotated<Processor>,
		Observable<ProcessorFinishedEvent>, WorkflowItem {
	/**
	 * The iteration strategy is responsible for combining input data events
	 * into jobs which are then queued for execution through the dispatch stack
	 * 
	 * @return IterationStrategyStack containing one or more IterationStrategy
	 *         objects. In most cases this will only contain a single
	 *         IterationStrategy but there are particular scenarios where
	 *         staging partial iteration strategies together is the only way to
	 *         get the desired combination of inputs
	 */
	@HierarchyTraversal(hierarchies = { "workflowStructure" }, role = { CHILD })
	IterationStrategyStack getIterationStrategy();

	/**
	 * Each processor has a list of zero or more input ports. These are uniquely
	 * named within the list. Any input port may have a default value associated
	 * with it and may be attached to exactly one upstream output port. Where it
	 * is necessary to connect a single input port to multiple output ports a
	 * Merge object is used. Ordering within the list is not meaningful but we
	 * use List rather than Set to preserve the ordering across serialisation
	 * operations.
	 * <p>
	 * Processor inputs are instances of FilteringInputPort - they must have the
	 * filter depth set before any data events arrive at the Processor. In
	 * addition they assume that a full collection will be supplied, i.e. that
	 * there will be exactly one event at the end of the list of events for a
	 * given process ID with an index array of length zero.
	 * 
	 * @return List of named input ports
	 */
	@Override
	@HierarchyTraversal(hierarchies = { "workflowStructure" }, role = { CHILD })
	List<? extends ProcessorInputPort> getInputPorts();

	/**
	 * Each processor has a list of zero or more output ports. Output ports are
	 * uniquely named within the list and may be connected to arbitrarily many
	 * downstream input ports or Merge objects. Ordering within the list is not
	 * meaningful but we use List rather than Set to preserve the ordering
	 * across serialisation operations.
	 * 
	 * @return List of named output ports
	 */
	@Override
	@HierarchyTraversal(hierarchies = { "workflowStructure" }, role = { CHILD })
	List<? extends ProcessorOutputPort> getOutputPorts();

	/**
	 * The dispatch stack pulls jobs from the queue generated by the iteration
	 * system and handles the dispatch of these jobs to appropriate activity
	 * workers
	 * 
	 * @return the DispatchStackImpl for this processor
	 */
	@HierarchyTraversal(hierarchies = { "workflowStructure" }, role = { CHILD })
	DispatchStack getDispatchStack();

	/**
	 * A processor contains zero or more activities in an ordered list. To be
	 * any use in a workflow the processor should contain at least one activity
	 * but it's technically valid to have none! Activities may be abstract or
	 * concrete where an abstract activity is one with no invocation mechanism,
	 * in these cases additional logic must be added to the dispatch stack of
	 * the containing processor to convert these to concrete invokable
	 * activities during the workflow invocation.
	 * 
	 * @return list of Activity instances
	 */
	@HierarchyTraversal(hierarchies = { "workflowStructure" }, role = { CHILD })
	List<? extends Activity<?>> getActivityList();

	/**
	 * A processor with no inputs cannot be driven by the supply of data tokens
	 * as it has nowhere to receive such tokens. This method allows a processor
	 * to fire on an empty input set, in this case the owning process identifier
	 * must be passed explicitly to the processor. Internally this pushes a
	 * single empty job event into the dispatch queue, bypassing the iteration
	 * logic (which is entirely predicated on the existence of input ports).
	 * Callers must ensure that an appropriate process identifier is specified,
	 * the behaviour on missing or duplicate process identifiers is not defined.
	 */
	void fire(String owningProcess, InvocationContext context);

	/**
	 * A processor has zero or more preconditions explicitly declared. All such
	 * preconditions must be satisfied before any jobs are passed into the
	 * dispatch stack. These preconditions replace and generalise the
	 * coordination constraints from Taverna 1.
	 * 
	 * @return a List of Condition objects defining constraints on this
	 *         processor's execution
	 */
	@HierarchyTraversal(hierarchies = { "workflowStructure" }, role = { CHILD })
	List<? extends Condition> getPreconditionList();

	/**
	 * A processor may control zero or more other processors within the same
	 * level of the workflow through preconditions.
	 * 
	 * @return a List of Condition objects for which this is the controlling
	 *         processor
	 */
	List<? extends Condition> getControlledPreconditionList();
}
